/* Copyright 2016 Bas Nieuwenhuizen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <algorithm>
#include <array>
#include <atomic>
#include <cassert>
#include <chrono>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <map>
#include <mutex>
#include <random>
#include <set>
#include <sstream>
#include <string>
#include <thread>
#include <vector>

#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <unordered_map>
#include <unordered_set>
#include <regex>

bool string_matches(const char *str1, const char *str2)
{
  return strncmp(str1, str2, strlen(str2)) == 0;
}

bool is_excluded_test(std::string const &testname,
                      const std::vector<std::regex> &excluded_tests)

{
  std::vector<std::regex>::const_iterator it;
  for (it = excluded_tests.begin(); it != excluded_tests.end(); ++it) {
    if (regex_match(testname, *it))
      return true;
  }
  return false;
}

std::vector<std::string> parse_testcase_file(std::string const &filename,
                                             const std::vector<std::regex> &excluded_tests) {
  std::vector<std::string> cases;
  std::ifstream in(filename);
  if (!in.is_open()) {
    std::cerr << "could not find testcase file \"" << filename << "\""
              << std::endl;
    std::exit(1);
  }

  std::string line;
  while (std::getline(in, line)) {
    if (!is_excluded_test(line, excluded_tests))
      cases.push_back(line);
  }
  return cases;
}

std::vector<std::string> get_testcases(std::string const &deqp,
                                       const std::vector<std::regex> &excluded_tests,
                                       const std::vector<std::string> &deqp_args) {
  std::string dir = std::filesystem::path(deqp).parent_path().native();
  FILE *f;
  char buf[4096];
  std::vector<std::string> cases;
  int fd[2];

  std::vector<std::string> args;
  args.push_back(deqp.c_str());
  args.push_back("--deqp-runmode=stdout-caselist");
  args.insert(args.end(), deqp_args.begin(), deqp_args.end());

  const char *c_args[args.size() + 1];
  for (unsigned i = 0; i < args.size(); i++) {
    c_args[i] = args[i].c_str();
  }
  c_args[args.size()] = NULL;

  if (pipe(fd))
    throw - 1;

  if (!fork()) {
    close(fd[0]);
    dup2(fd[1], 1);
    dup2(fd[1], 2);
    if (chdir(dir.c_str()))
      throw - 1;
    execv(deqp.c_str(), (char *const *)c_args);
  }
  close(fd[1]);
  f = fdopen(fd[0], "r");

  while (fgets(buf, 4096, f)) {
    if (string_matches(buf, "TEST: ")) {
      auto len = strlen(buf);
      std::string testname = std::string(buf + 6, buf + len - 1);

      if (!is_excluded_test(testname, excluded_tests))
        cases.push_back(testname);
    } else if (string_matches(buf, "FATAL ERROR: ")) {
      std::cerr << "Error getting caselist:\n";
      std::cerr << buf;
    }
  }
  fclose(f);
  return cases;
}

enum status {
  PASS,
  FAIL,
  SKIP,
  XFAIL,
  UNEXPECTEDPASS,
  CRASH,
  TIMEOUT,
  MISSING,
  FLAKE,
};
#define STATUS_COUNT ((int)FLAKE + 1)

std::string get_status_name(enum status status)
{
  switch (status) {
  case PASS:
    return "Pass";
  case FAIL:
    return "Fail";
  case SKIP:
    return "Skip";
  case XFAIL:
    return "ExpectedFail";
  case UNEXPECTEDPASS:
    return "UnexpectedPass";
  case CRASH:
    return "Crash";
  case TIMEOUT:
    return "Timeout";
  case MISSING:
    return "Missing";
  case FLAKE:
    return "Flake";
  default:
    abort();
  }
}

struct Context {
  unsigned device_id;
  std::string deqp;
  std::vector<std::string> test_cases;
  std::chrono::time_point<std::chrono::steady_clock> start_time;
  bool compact_display;

  std::map<std::string, enum status> results;
  std::mutex results_mutex;

  std::atomic<std::size_t> taken_cases, finished_cases;
  double timeout = 60.0;
  std::vector<std::string> deqp_args;

  std::atomic<std::size_t> status_counts[STATUS_COUNT];

  /* Set of tests supplied by the user as expected fails.  Failures
   * and crashes of those tests will be logged as fails (and allow the
   * test suite as a whole to return success), while passes of those
   * tests will be logged as unexpected passes (and cause the test
   * suite as a whole to return failure, go update your xfail list).
   */
  std::unordered_set<std::string> xfails;

  Context() {
    for (int i = 0; i < STATUS_COUNT; i++)
      status_counts[i] = 0;
  }
};

class Line_reader {
public:
  ~Line_reader();

  void set_fd(int fd);

  enum read_status {
    SUCCESS,
    PARTIAL,
    END,
    TIMEOUT,
    FAIL

  };

  read_status read(char **text, double timeout);

private:
  int fd_ = -1;
  std::array<char, 4096> buf_;
  int valid_sz_ = 0;
  int skip_sz_ = 0;
};

Line_reader::~Line_reader() {
  if (fd_ >= 0)
    close(fd_);
}

void Line_reader::set_fd(int fd) {
  if (fd_ >= 0)
    close(fd_);
  fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);
  fd_ = fd;
  valid_sz_ = 0;
  skip_sz_ = 0;
}

Line_reader::read_status Line_reader::read(char **text, double timeout) {
  assert(fd_ >= 0);

  if (skip_sz_) {
    valid_sz_ -= skip_sz_;
    memmove(buf_.data(), buf_.data() + skip_sz_, valid_sz_);
    skip_sz_ = 0;
  }

  char *newline;
  bool end = false;
  while (!(newline = (char *)memchr(buf_.data(), '\n', valid_sz_)) &&
         !end && valid_sz_ + 1 < (int)buf_.size()) {
    for (;;) {
      struct pollfd p = {};
      p.fd = fd_;
      p.events = POLLIN | POLLHUP;

      int r = poll(&p, 1, (int)(timeout * 1000.0));
      if (r > 0)
        break;
      else if (r == 0)
        return TIMEOUT;
      else if (r < 0 && (errno == EINTR || errno == EAGAIN))
        continue;
      else
        return FAIL;
    }
    for (;;) {
      ssize_t r =
          ::read(fd_, buf_.data() + valid_sz_, buf_.size() - 1 - valid_sz_);
      if (r >= 0) {
        valid_sz_ += r;
        end = r == 0;
        break;
      } else if (r < 0 && errno == EAGAIN)
        break;
      else if (r < 0 && errno != EINTR)
        return FAIL;
    }
  }

  if (newline) {
    skip_sz_ = newline - buf_.data() + 1;
    *newline = 0;
    *text = buf_.data();
    return SUCCESS;
  } else {
    buf_[valid_sz_] = 0;
    skip_sz_ = valid_sz_;
    *text = buf_.data();
    return end ? END : PARTIAL;
  }
}

class ProcessBlockState {
public:
  const Context &ctx;
  std::set<std::string> tests;
  std::map<std::string, enum status> *results;

  bool start; /* Set if we are due to spawn a new dEQP instance. */
  bool test_active; /* Set if we're between test start and test status */

  std::string active_test;

  void record_result(enum status status) {
    assert(test_active);

    if (ctx.xfails.find(active_test) != ctx.xfails.end()) {
      switch (status) {
      case FAIL:
      case CRASH:
        status = XFAIL;
        break;
      case PASS:
        status = UNEXPECTEDPASS;
        break;
      default:
        break;
      }
    }

    results->insert({active_test, status});
    test_active = false;
  }

  ProcessBlockState(const Context &ctx, const std::set<std::string> &tests, std::map<std::string, enum status> *results)
    : ctx(ctx), tests(tests), results(results)
  {
    start = true;
    test_active = false;
  }
};



std::map<std::string, enum status>
run_block(const Context &ctx, unsigned thread_id, unsigned iteration,
          const std::set<std::string>& tests) {

  std::string filename = "";
  std::string qpafilename = "";
  Line_reader reader;
  bool before_first_test = true;
  std::map<std::string, enum status> results;
  bool has_fails = false;
  int idx = 0;

  ProcessBlockState state(ctx, tests, &results);

  /* Loop running the test suite on the remaining tests to run until
   * we have processed them all.  This may take more than one run in
   * the case of crashes.
   */
  while (state.test_active || !state.tests.empty()) {
    if (state.start) {
      if ((filename != "") && !has_fails) {
        unlink(filename.c_str());
        unlink(qpafilename.c_str());
      }
      filename = "/tmp/deqp_runner." + std::to_string(getpid()) + "." +
          std::to_string(thread_id) + "." +
          std::to_string(iteration) + "." +
          std::to_string(idx++) + ".txt";
      qpafilename = "/tmp/deqp_runner." + std::to_string(getpid()) + "." +
          std::to_string(thread_id) + "." +
          std::to_string(iteration) + "." +
          std::to_string(idx++) + ".qpa";
      std::string dir = std::filesystem::path(ctx.deqp).parent_path().native();

      int fd[2];
      std::ofstream out(filename);
      for (auto &&e : state.tests)
        out << e << "\n";
      out.close();

      std::vector<std::string> args;
      args.push_back(ctx.deqp.c_str());
      args.push_back("--deqp-log-flush=disable");
      args.push_back("--deqp-log-filename=" + qpafilename);
      args.push_back("--deqp-caselist-file=" + filename);
      args.push_back("--deqp-vk-device-id=" + std::to_string(ctx.device_id));
      args.insert(args.end(), ctx.deqp_args.begin(), ctx.deqp_args.end());

      const char *c_args[args.size() + 1];
      for (unsigned i = 0; i < args.size(); i++) {
        c_args[i] = args[i].c_str();
      }
      c_args[args.size()] = NULL;

      if (pipe2(fd, O_CLOEXEC))
        throw - 1;

      if (!fork()) {
        dup2(fd[1], 1);
        dup2(fd[1], 2);
        if (chdir(dir.c_str()))
          throw - 1;

        execv(ctx.deqp.c_str(), (char *const *)c_args);
      }
      close(fd[1]);
      reader.set_fd(fd[0]);
      state.start = false;
      state.test_active = false;
      before_first_test = true;
      has_fails = false;
    }
    char *line = NULL;
    auto r = reader.read(&line, ctx.timeout);

    if (r == Line_reader::FAIL) {
      abort();
    } else if (r == Line_reader::TIMEOUT) {
      state.start = true;
      if (state.test_active) {
        state.record_result(TIMEOUT);
        has_fails = true;
      }
    } else if (r == Line_reader::END || string_matches(line, "DONE!")) {
      state.start = true;
      if (state.test_active) {
        state.record_result(CRASH);
        has_fails = true;
      } else if(before_first_test) {
        while(!state.tests.empty()) {
          auto it = state.tests.begin();
          state.active_test = *it;
          state.test_active = true;
          state.tests.erase(it);

          state.record_result(MISSING);
        }
      }
    } else if (string_matches(line, "  NotSupported")) {
      state.record_result(SKIP);
    } else if (string_matches(line, "  Fail") ||
               string_matches(line, "  InternalError")) {
      state.record_result(FAIL);
      has_fails = true;
    } else if (string_matches(line, "  Pass") ||
               string_matches(line, "  CompatibilityWarning") ||
               string_matches(line, "  QualityWarning")) {
      state.record_result(PASS);
    } else if (string_matches(line, "Test case '")) {
      auto len = strlen(line) - 3;
      auto name = std::string(line + 11, line + len);
      auto it = state.tests.find(name);
      if (it == state.tests.end()) {
        abort();
      }
      state.active_test = name;
      state.tests.erase(it);
      state.test_active = true;
      before_first_test = false;
    } else if (string_matches(line, "FATAL ERROR: ")) {
      if (state.test_active) {
        state.record_result(FAIL);
        has_fails = true;
      } else {
        std::cerr << line;
        std::cerr << "\n";
        abort();
      }
    }
  }
  if ((filename != "") && !has_fails) {
    unlink(filename.c_str());
    unlink(qpafilename.c_str());
  }
  return results;
}

std::set<std::string> grab_tests(Context &ctx) {
  /* Use cmpxchg to take a block of tests from the whole list */
  std::size_t base_idx, count;
  do {
    base_idx = ctx.taken_cases.load();
    if (base_idx >= ctx.test_cases.size())
      return {};
    count = std::min<std::size_t>(
        std::max<std::size_t>((ctx.test_cases.size() - base_idx) / 32, 1), 128);

  } while (
      !ctx.taken_cases.compare_exchange_strong(base_idx, base_idx + count));

  std::set<std::string> tests;
  for (std::size_t i = 0; i < count; ++i)
    tests.insert(ctx.test_cases[base_idx + i]);
  return tests;
}

void update_results(Context &ctx, std::map<std::string, enum status> results) {
  std::lock_guard<std::mutex> lck (ctx.results_mutex);
  ctx.results.insert(results.begin(), results.end());
  for (auto&& e : results)
    ctx.status_counts[e.second]++;

  ctx.finished_cases += results.size();
}

void test_for_flakes(const Context &ctx, unsigned thread_id, unsigned iteration,
                     const std::set<std::string>& tests,
                     std::map<std::string, enum status> *results) {
  bool has_failures = false;
  for (auto&& e : *results) {
    switch(e.second) {
    case FAIL:
    case UNEXPECTEDPASS:
    case CRASH:
    case TIMEOUT:
      has_failures = true;
      break;
    default:
      break;
    }
  }

  if (!has_failures)
    return;

  auto retry_results = run_block(ctx, thread_id, iteration, tests);

  assert(retry_results.size() == results->size());
  for (auto& e : *results) {
    auto it = retry_results.find(e.first);

    assert (it != retry_results.end());

    if (e.second != it->second)
      e.second = FLAKE;
  }
}

bool process_block(Context &context, unsigned thread_id, unsigned iteration) {
  auto tests = grab_tests(context);
  if (tests.empty())
    return false;

  auto results = run_block(context, thread_id, iteration, tests);

  assert(tests.size() == results.size());

  test_for_flakes(context, thread_id, iteration + 1, tests, &results);

  update_results(context, std::move(results));
  return true;
}

void thread_runner(Context &ctx, unsigned thread_id) {
  unsigned iteration = 0;
  /* increment iteration by 2 each time through, the odd iteration #s
   * are (optionally) used to re-run in case of flakes
   */
  while (process_block(ctx, thread_id, iteration))
    iteration += 2;
}

std::string format_duration(std::int64_t seconds) {
  std::string ret;
  if (seconds >= 3600) {
    ret += std::to_string(seconds / 3600);
    ret += ':';
    seconds %= 3600;
  }

  if (seconds >= 60 || !ret.empty()) {
    if (!ret.empty() && seconds < 600)
      ret += '0';
    ret += std::to_string(seconds / 60);
    ret += ':';
    seconds %= 60;
  }
  if (!ret.empty() && seconds < 10)
    ret += '0';
  ret += std::to_string(seconds);
  return ret;
}

void update(Context &ctx) {
  std::size_t status_counts[STATUS_COUNT];
  /* Snapshot the status counts from the context, so our numbers in
   * the report add up even if the atomics are being updated.
   */
  std::size_t total = 0;
  for (int i = 0; i < STATUS_COUNT; i++) {
    status_counts[i] = ctx.status_counts[i];
    total += status_counts[i];
  }

  std::chrono::time_point<std::chrono::steady_clock> current_time =
      std::chrono::steady_clock::now();
  auto duration = std::chrono::duration_cast<std::chrono::duration<double>>(
      current_time - ctx.start_time);

  std::cout << "[" << total << "/" << ctx.test_cases.size() << "]";
  for (int i = 0; i < STATUS_COUNT; i++) {
    std::cout << " "
              << get_status_name((enum status)i)
              << ": "
              << status_counts[i];
  }
  std::cout << " Duration: " << format_duration(duration.count());
  if (total) {
    std::cout << " Remaining: "
              << format_duration(duration.count() / total *
                                 (ctx.test_cases.size() - total));
  }
  if (ctx.compact_display)
          std::cout << "                  \r";
  else
          std::cout << "\n";
  std::cout.flush();
}

void
usage(char *progname)
{
  std::cerr << "Usage: " << progname << "\n";
  std::cerr << "    --deqp <deqp-vk or deqp-gles* binary>\n";
  std::cerr << "    --output <log filename>\n";
  std::cerr << "    [--caselist <mustpass.txt>]\n";
  std::cerr << "    [--xfail-list <caselist.txt>]\n";
  std::cerr << "    [--exclude-list <caselist.txt>]\n";
  std::cerr << "    [--timeout <seconds>]\n";
  std::cerr << "    [--exclude-tests <regex,regex,...>]\n";
  std::cerr << "    [--job <threadcount>]\n";
  std::cerr << "    [--device <vk device id>]\n";
  std::cerr << "    [--allow-flakes true/false]\n";
  std::cerr << "    [--compact-display true/false]\n";
  std::cerr << "    [-- --deqp-binary-arguments]\n";
  std::cerr << "\n";

  std::cerr << "--timeout defaults to 60 seconds of no new dEQP output\n";
  std::cerr << "--job defaults to CPU thread count\n";
  exit(1);
}

std::map<std::string, std::string> parse_args(int argc, char *argv[],
					      std::vector<std::string> *deqp_args) {
  std::map<std::string, std::string> args;
  for (int i = 1; i < argc; ++i) {
    if (argv[i][0] != '-' || argv[i][1] != '-') {
      std::cerr << "arguments have to start with '--'\n";
      usage(argv[0]);
    }
    auto eq = strchr(argv[i], '=');
    if (eq) {
      std::string arg_name{argv[i] + 2, eq};
      std::string value = eq + 1;
      if (args.find(arg_name) != args.end()) {
        std::cerr << "specified '" << arg_name << "' twice.\n";
	usage(argv[0]);
      }
      args[arg_name] = value;
    } else if (strcmp(argv[i], "--") != 0) {
      std::string arg_name = argv[i] + 2;
      ++i;
      if (i == argc) {
        std::cerr << "missing value for argument '" << arg_name << "'\n";
	usage(argv[0]);
      }
      if (args.find(arg_name) != args.end()) {
        std::cerr << "specified '" << arg_name << "' twice.\n";
	usage(argv[0]);
      }
      args[arg_name] = argv[i];
    } else {
      ++i;
      for (; i < argc; ++i) {
	deqp_args->push_back(argv[i]);
      }
    }
  }
  return args;
}

int main(int argc, char *argv[]) {
  signal(SIGCHLD, SIG_IGN);

  std::vector<std::string> deqp_args;
  auto args = parse_args(argc, argv, &deqp_args);
  if (args.find("deqp") == args.end()) {
    std::cerr << "--deqp missing\n";
    usage(argv[0]);
  }
  if (args.find("output") == args.end()) {
    std::cerr << "--output missing\n";
    usage(argv[0]);
  }

  std::vector<std::regex> excluded_tests;
  if (args.find("exclude-tests") != args.end()) {
    std::istringstream excluded(args.find("exclude-tests")->second);

    while (excluded.good()) {
      std::string test;
      getline(excluded, test, ',');
      excluded_tests.push_back(std::regex(test));
    }
  }

  if (args.find("exclude-list") != args.end()) {
    std::string filename = args.find("exclude-list")->second;
    std::ifstream in(filename);
    if (!in.is_open()) {
      std::cerr << "could not find exclude file \"" << filename << "\""
                << std::endl;
      std::exit(1);
    }

    std::string line;
    while (std::getline(in, line)) {
      if (line.size() > 0 && line[0] != '#')
        excluded_tests.push_back(std::regex(line));
    }
  }

  Context ctx;
  ctx.deqp = args.find("deqp")->second;
  ctx.deqp_args = deqp_args;
  if (args.find("caselist") == args.end()) {
    ctx.test_cases = get_testcases(ctx.deqp, excluded_tests, deqp_args);
  } else {
    ctx.test_cases = parse_testcase_file(args.find("caselist")->second,
                                         excluded_tests);
  }

  if (args.find("xfail-list") != args.end()) {
    std::string filename = args.find("xfail-list")->second;
    std::ifstream in(filename);
    if (!in.is_open()) {
      std::cerr << "could not find xfail file \"" << filename << "\""
                << std::endl;
      std::exit(1);
    }

    std::string line;
    while (std::getline(in, line))
      ctx.xfails.insert(line);
  }

  if (args.find("timeout") != args.end())
    ctx.timeout = strtof(args.find("timeout")->second.c_str(), NULL);

  int job = 0;
  if (args.find("job") != args.end())
    job = strtol(args.find("job")->second.c_str(), NULL, 10);

  ctx.device_id = 1;
  if (args.find("device") != args.end())
    ctx.device_id = strtol(args.find("device")->second.c_str(), NULL, 10);

  if (args.find("compact-display") == args.end() || args.find("compact-display")->second == "true")
    ctx.compact_display = true;
  else
    ctx.compact_display = false;

  if (args.find("shuffle") == args.end() || args.find("shuffle")->second == "true") {
          /* Shuffle the tests so that the groups that our threads grab are
           * more or less evenly distributed in test namespace (since test
           * runtime is closely related to test name).
           */
          std::mt19937 rng;
          std::shuffle(ctx.test_cases.begin(), ctx.test_cases.end(), rng);
  }

  auto thread_count = job > 0 ? job : std::thread::hardware_concurrency();
  ctx.taken_cases = 0;
  ctx.finished_cases = 0;
  ctx.start_time = std::chrono::steady_clock::now();
  std::vector<std::thread> threads;
  for (unsigned i = 0; i < thread_count; ++i) {
    threads.emplace_back([&ctx, i]() { thread_runner(ctx, i); });
  }

  while (ctx.finished_cases.load() < ctx.test_cases.size()) {
    update(ctx);
    if (ctx.compact_display)
            sleep(1);
    else
            sleep(2);
  }
  update(ctx);

  std::cout << "\n";

  for (auto &t : threads)
    t.join();

  assert(ctx.test_cases.size() == ctx.results.size());

  std::ofstream out(args.find("output")->second);
  for (auto &&entry : ctx.results)
    out << entry.first << "," << get_status_name(entry.second) << "\n";
  update(ctx);
  std::cout << "\n";

  for (int i = 0; i < STATUS_COUNT; i++) {
    switch ((enum status)i) {
    case PASS:
    case SKIP:
    case XFAIL:
      break;
    case FLAKE:
      if (args.find("allow-flakes") != args.end() && args.find("allow-flakes")->second == "true")
        break;
      /* Fallthrough */
    default:
      if (ctx.status_counts[i] != 0)
        return 1;
    }
  }

  return 0;
}
